# Copyright (C) 2012 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Main gwibber-service module.

This gets turned into a script by `python3 setup.py install`.
"""

__all__ = [
    'main',
    ]


import logging

from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import Gio, GLib

from gwibber.service.connection import ConnectionMonitor
from gwibber.service.dispatcher import Dispatcher
from gwibber.service.shortener import URLShorten
from gwibber.utils.logging import initialize
from gwibber.utils.menus import MenuManager
from gwibber.utils.options import Options


# Logger must be initialized before it can be used.
log = None


def main():
    global log
    # Initialize command line options.
    args = Options().parser.parse_args()
    if args.list_protocols:
        from gwibber.utils.manager import protocol_manager
        for name in sorted(protocol_manager.protocols):
            cls = protocol_manager.protocols[name]
            # Stub protocols have no version so don't print them.
            version = getattr(cls.info, 'version', None)
            if version is not None:
                print(cls.__name__, 'version', version)
        return
    # Initialize the logging subsystem.
    gsettings = Gio.Settings.new('org.gwibber.preferences')
    initialize(console=args.console,
               debug=args.debug or gsettings.get_boolean('debug'))
    log = logging.getLogger('gwibber.service')
    log.info('Gwibber backend service starting')

    # Set up the DBus main loop.
    DBusGMainLoop(set_as_default=True)
    loop = GLib.MainLoop()

    refresh_interval = max(gsettings.get_int('interval'), 5) * 60

    # Load up the various services.  We do it this way so that we retain
    # references to the service endpoints without pyflakes screaming at us
    # about unused local variables.
    class services:
        connection = ConnectionMonitor()
        dispatcher = Dispatcher(loop, refresh_interval)
        menus = MenuManager(dispatcher.Refresh, loop.quit)
        shorten = URLShorten(gsettings)

    if args.test:
        # This module is only necessary if we're running the unit tests.
        from gwibber.testing.service import TestService
        services.test = TestService(services.connection)
    try:
        log.info('Starting gwibber-service main loop')
        loop.run()
    except KeyboardInterrupt:
        log.info('Stopped gwibber-service main loop')
