# Copyright (C) 2012 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The Facebook protocol plugin."""

__all__ = [
    'Facebook',
    ]


import logging

from datetime import datetime, timedelta

from gwibber.utils.base import Base, feature
from gwibber.utils.download import get_json
from gwibber.utils.time import parsetime, iso8601utc


# 'id' can be the id of *any* facebook object
# https://developers.facebook.com/docs/reference/api/
API_BASE = 'https://graph.facebook.com/{id}'
ME_URL = API_BASE.format(id='me')
LIMIT = 100

log = logging.getLogger('gwibber.service')


class Facebook(Base):
    def _whoami(self, authdata):
        """Identify the authenticating user."""
        me_data = get_json(
            ME_URL, dict(access_token=self._account.access_token))
        self._account.user_id = me_data.get('id')
        self._account.user_name = me_data.get('name')

    def _is_error(self, data):
        """Is the return data an error response?"""
        error = data.get('error')
        if error is None:
            return False
        log.error('Facebook error ({} {}): {}'.format(
            error.get('code'), error.get('type'), error.get('message')))
        return True

    def _publish_entry(self, entry):
        message_id = entry.get('id')
        if message_id is None:
            # We can't do much with this entry.
            return

        args = dict(stream='messages')
        from_record = entry.get('from')
        if from_record is not None:
            args['sender'] = sender_id = from_record.get('id', '')
            args['sender_nick'] = from_record.get('name', '')
            args['from_me'] = (sender_id == self._account.user_id)
        # Normalize the timestamp.
        timestamp = entry.get('updated_time', entry.get('created_time'))
        if timestamp is not None:
            args['timestamp'] = iso8601utc(parsetime(timestamp))
        args['message'] = entry.get('message', '')
        args['url'] = API_BASE.format(id=message_id)
        args['icon_uri'] = entry.get('icon', '')
        like_count = entry.get('likes', {}).get('count')
        if like_count is not None:
            args['likes'] = like_count
            args['liked'] = (like_count > 0)
        args['link_picture'] = entry.get('picture', '')
        args['link_name'] = entry.get('name', '')
        args['link_url'] = entry.get('link', '')
        args['link_desc'] = entry.get('description', '')
        args['link_caption'] = entry.get('caption', '')
        # Parse comments now.
        all_comments = []
        for comment_data in entry.get('comments', {}).get('data', []):
            comment_message = comment_data.get('message')
            if comment_message is not None:
                all_comments.append(comment_message)
        args['comments'] = all_comments
        self._publish(message_id, **args)

    @feature
    def receive(self, since=None):
        """Retrieve a list of Facebook objects.

        A maximum of 100 objects are requested.

        :param since: Only get objects posted since this date.  If not given,
            then only objects younger than 10 days are retrieved.  The value
            is a number seconds since the epoch.
        :type since: float
        """
        access_token = self._get_access_token()
        if since is None:
            when = datetime.now() - timedelta(days=10)
        else:
            when = datetime.fromtimestamp(since)
        entries = []
        url = ME_URL + '/home'
        params = dict(access_token=access_token,
                      since=when.isoformat(),
                      limit=LIMIT)
        # Now access Facebook and follow pagination until we have at least
        # LIMIT number of entries, or we've reached the end of pages.
        while True:
            response = get_json(url, params)
            if self._is_error(response):
                # We'll just use what we have so far, if anything.
                break
            data = response.get('data')
            if data is None:
                # I guess we're done.
                break
            entries.extend(data)
            if len(entries) >= LIMIT:
                break
            # We haven't gotten the requested number of entries.  Follow the
            # next page if there is one to try to get more.
            pages = response.get('paging')
            if pages is None:
                break
            # The 'next' key has the full link to follow; no additional
            # parameters are needed.  Specifically, this link will already
            # include the access_token, and any since/limit values.
            url = pages.get('next')
            params = None
            if url is None:
                # I guess there are no more next pages.
                break
        # We've gotten everything Facebook is going to give us.  Now, decipher
        # the data and publish it.
        # https://developers.facebook.com/docs/reference/api/post/
        for entry in entries:
            self._publish_entry(entry)

    @feature
    def like(self, obj_id):
        """Like any arbitrary object on Facebook.

        This includes messages, statuses, wall posts, events, etc.
        """
        url = API_BASE.format(id=obj_id) + '/likes'
        token = self._get_access_token()

        if not get_json(url, post=True, params=dict(access_token=token)):
            log.error('Failed to like {} on Facebook'.format(obj_id))

    @feature
    def send(self, message, obj_id='me'):
        """Write a message on somebody or something's wall.

        If you don't specify whose, it defaults to yours. obj_id can
        be literally anything that has a wall, be it a user, an app, a
        company, an event, etc.
        """
        url = API_BASE.format(id=obj_id) + '/feed'
        token = self._get_access_token()

        result = get_json(url, post=True, params=dict(access_token=token,
                                                      message=message))
        new_id = result.get('id')
        if new_id is None:
            log.error('Failed sending to Facebook: {!r}'.format(result))
            return

        url = API_BASE.format(id=new_id)
        entry = get_json(url, params=dict(access_token=token))
        self._publish_entry(entry)
