# Copyright (C) 2012 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""DBus service object for URL shortening."""

__all__ = [
    'URLShorten',
    ]


import logging

import dbus
import dbus.service

from gwibber.shorteners import lookup

log = logging.getLogger('gwibber.service')


class URLShorten(dbus.service.Object):
    __dbus_object_path__ = '/com/gwibber/URLShorten'

    def __init__(self, settings):
        self.settings = settings
        self.bus = dbus.SessionBus()
        bus_name = dbus.service.BusName("com.Gwibber.URLShorten", bus=self.bus)
        super().__init__(bus_name, self.__dbus_object_path__)

    @dbus.service.method("com.Gwibber.URLShorten",
                         in_signature="s", out_signature="s")
    def Shorten(self, url):
        """Shorten a URL.

        Takes a url as a string and returns a shortened url as a string.

        >>> import dbus
        >>> url = 'http://www.example.com/this/is/a/long/url'
        >>> obj = dbus.SessionBus().get_object(
        ...     'com.Gwibber.URLShorten',
        ...     '/com/gwibber/URLShorten')
        >>> shortener = dbus.Interface(obj, 'com.Gwibber.URLShorten')
        ... short_url = shortener.Shorten(url)
        """
        service_name = self.settings.get_string('urlshorter')
        log.info("Shortening URL %s with %s", url, service_name)
        if (lookup.is_shortened(url) or
            not self.settings.get_boolean('shorten-urls')):
            # It's already shortened, or the preference is not set.
            return url
        service = lookup.lookup(service_name)
        try:
            return service.shorten(url)
        except Exception:
            log.exception('URL shortening class: {}'.format(service))
            return url
