# Copyright (C) 2012 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Convenient downloading."""

__all__ = [
    'Downloader',
    'RateLimiter',
    'get_json',
    ]


import json
import logging

from base64 import encodebytes
from contextlib import contextmanager
from urllib.parse import urlencode
from urllib.request import Request, urlopen


log = logging.getLogger('gwibber.service')


class RateLimiter:
    """Base class for the rate limiting API.

    By default, this class does no rate limiting.  Subclass from this and
    override the `wait()` and `update()` methods for protocol specific
    rate-limiting functionality.
    """
    def wait(self, request):
        """Wait an appropriate amount of time before returning.

        Downloading is blocked until this method returns.  This does not block
        the entire application since downloading always happens in a
        sub-thread.  If no wait is necessary, return immediately.

        :param request: The soon-to-be executed url request.
        :type request: urllib.request.Request
        """
        pass

    def update(self, response):
        """Update any rate limiting values based on the service's response.

        :param response: A urllib response object.
        """
        pass


class Downloader:
    """Convenient downloading wrapper."""

    def __init__(self, url, params=None, post=False,
                 username=None, password=None,
                 headers=None, rate_limiter=None):
        self.url = url
        self.params = params
        self.post = post
        self.username = username
        self.password = password
        self.headers = ({} if headers is None else headers)
        self._rate_limiter = (RateLimiter() if rate_limiter is None
                              else rate_limiter)
        # XXX Proxy handling.

    def _build_request(self):
        """Return a request object, with all the right headers.

        :return: The request.
        :rtype: urllib.request.Request
        """
        data = None
        url = self.url
        headers = self.headers.copy()
        if self.params is not None:
            # urlencode() does not have an option to use quote() instead of
            # quote_plus(), but Twitter requires percent-encoded spaces, and
            # this is harmless to any other protocol.
            params = urlencode(self.params).replace('+', '%20')
            if self.post:
                data = params.encode('utf-8')
                headers['Content-Type'] = (
                    'application/x-www-form-urlencoded; charset=utf-8')
            else:
                # Do a GET with an encoded query string.
                url = '{}?{}'.format(self.url, params)
        request = Request(url, data, headers)
        if self.username is not None and self.password is not None:
            auth = '{}:{}'.format(self.username, self.password).encode('utf-8')
            # encodebytes() includes a bogus trailing newline, which we must
            # strip off.
            value = encodebytes(auth)[:-1].decode('utf-8')
            basic = 'Basic {}'.format(value)
            request.add_header('Authorization', basic)
        # Possibly do some rate limiting.
        self._rate_limiter.wait(request)
        return request

    @contextmanager
    def _download(self):
        """Perform the download, as a context manager."""
        request = self._build_request()
        with urlopen(request) as result:
            yield result
            self._rate_limiter.update(result)

    def get_bytes(self):
        """Return the results as a bytes object."""
        with self._download() as result:
            return result.read()

    def get_string(self):
        """Return the results as a string, decoded as per the response."""
        with self._download() as result:
            payload = result.read()
            encoding = result.info().get_content_charset()
            return payload.decode(encoding)

    def get_json(self):
        """Interpret and return the results as JSON data."""
        with self._download() as result:
            payload = result.read()
            info = result.info()
        # RFC 4627 $3.  JSON text SHALL be encoded in Unicode.  The default
        # encoding is UTF-8.  Since the first two characters of a JSON text
        # will always be ASCII characters [RFC0020], it is possible to
        # determine whether an octet stream is UTF-8, UTF-16 (BE or LE), or
        # UTF-32 (BE or LE) by looking at the pattern of nulls in the first
        # four octets.
        charset = info.get_content_charset()
        if charset is None:
            octet_0, octet_1, octet_2, octet_3 = payload[:4]
            if 0 not in (octet_0, octet_1, octet_2, octet_3):
                charset = 'utf-8'
            elif (octet_1 == octet_3 == 0) and octet_2 != 0:
                charset = 'utf-16le'
            elif (octet_0 == octet_2 == 0) and octet_1 != 0:
                charset = 'utf-16be'
            elif (octet_1 == octet_2 == octet_3 == 0):
                charset = 'utf-32le'
            elif (octet_0 == octet_1 == octet_2 == 0):
                charset = 'utf-32be'
        return json.loads(payload.decode(charset))


def get_json(*args, **kws):
    """Convenience method for most common use case."""
    return Downloader(*args, **kws).get_json()
